from collections.abc import Callable
from datetime import datetime

from api.miscellaneous import *
from api.objects.changelog import ChangelogVersionBuilder, ChangelogEntryError, ChangelogEntryCodec, SimpleChangelogEntrySubject
from api.objects.changelog_entries import ModificationEntry, DeletionEntry
from api.objects.field import ObjectField
from api.objects.objects import *
from api.objects.type import SEMESTER_STRING_PATTERN


def _encode_identity(value_json: str) -> str:
    if value_json is None:
        raise ChangelogEntryError("Cannot encode null values")
    return value_json


def _decode_legacy_boolean(legacy_value: str) -> bool:
    if legacy_value in ["1", "true", "True", "TRUE"]:
        return True
    if legacy_value in ["0", "false", "False", "FALSE"]:
        return False
    raise ChangelogEntryError(f"Unable to decode boolean '{truncate_string(legacy_value)}'")


def _encode_legacy_boolean(boolean_json: bool) -> str:
    if boolean_json is None:
        raise ChangelogEntryError("Cannot encode null boolean")
    return "1" if boolean_json else "0"


def _decode_legacy_shortstring_to_string_short(legacy_value: str) -> str:
    if len(legacy_value) > TYPE_STRING_SHORT.get_max_string_length():
        raise ChangelogEntryError(f"Legacy short string '{truncate_string(legacy_value)}' is too long")
    return legacy_value


def _decode_legacy_text_to_string_long(legacy_value: str) -> str:
    if len(legacy_value) > TYPE_STRING_LONG.get_max_string_length():
        raise ChangelogEntryError(f"Legacy text '{truncate_string(legacy_value)}' is too long")
    return legacy_value


# If this looks familiar, yes its copied and modified a bit from the sqlite_connector
_LEGACY_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"


def _decode_legacy_datetime(legacy_value: str) -> str:
    datetime_string = legacy_value.split(".")[0]  # Datetime sometimes contains a trailing .1234 for milliseconds, etc.
    try:
        return datetime.strptime(datetime_string, _LEGACY_DATETIME_FORMAT).strftime(API_DATETIME_FORMAT)
    except ValueError as e:
        raise ChangelogEntryError(f"Unable to decode legacy datetime '{truncate_string(legacy_value)}'") from e


def _encode_legacy_datetime(datetime_json: str) -> str:
    if datetime_json is None:
        raise ChangelogEntryError("Cannot encode null datetime")
    try:
        return datetime.strptime(datetime_json, API_DATETIME_FORMAT).strftime(_LEGACY_DATETIME_FORMAT)
    except ValueError as e:
        raise ChangelogEntryError(f"Unable to encode json datetime '{truncate_string(datetime_json)}'") from e


def _decode_legacy_integer_non_negative(legacy_value: str) -> int:
    try:
        value = int(legacy_value)
    except ValueError as e:
        raise ChangelogEntryError(f"Unable to decode legacy integer '{truncate_string(legacy_value)}'") from e
    if value < 0 or value > MAX_VALUE_SINT32:
        raise ChangelogEntryError(f"Legacy integer '{truncate_string(legacy_value)}' out of range")
    return value


def _encode_legacy_integer(value_json: int) -> str:
    if value_json is None:
        raise ChangelogEntryError("Cannot encode null integer")
    return str(value_json)


def _decode_legacy_string_to_id_string(legacy_value: str) -> str:
    if not (1 <= len(legacy_value) <= 100) or ID_STRING_PATTERN_NO_LENGTH.fullmatch(legacy_value) is None:
        raise ChangelogEntryError(f"Legacy string is not an id string: '{truncate_string(legacy_value)}'")
    return legacy_value


def _decode_legacy_string_to_semester_string(legacy_value: str) -> str:
    if legacy_value is None:
        raise ChangelogEntryError("Cannot encode null values")
    if len(legacy_value) == 0 or legacy_value.isspace():
        return "none"
    if len(legacy_value) > 100 or SEMESTER_STRING_PATTERN.fullmatch(legacy_value) is None:
        raise ChangelogEntryError(f"Legacy string is not a semester string: '{truncate_string(legacy_value)}'")
    return legacy_value


def _encode_legacy_semester_string(value_json: str) -> str:
    if value_json is None:
        raise ChangelogEntryError("Cannot encode null values")
    if value_json == "none":
        return ""
    return value_json


def _decode_legacy_announcement_level(legacy_value: str) -> str:
    level = _decode_legacy_integer_non_negative(legacy_value)
    match level:
        case 0: return "info_main_page"
        case 1: return "info_all_pages"
        case 2: return "warning_all_pages"
        case 3: return "important_all_pages_embed"
        case _: raise ChangelogEntryError(f"Unknown announcement level {level}")


def _encode_legacy_announcement_level(value_json: str) -> str:
    match value_json:
        case "info_main_page": return "0"
        case "info_all_pages": return "1"
        case "warning_all_pages": return "2"
        case "important_all_pages_embed": return "3"
        case _: raise ChangelogEntryError(f"Unknown announcement level {value_json}")


class _LegacyModificationEntryCodec(ChangelogEntryCodec[ObjectField]):
    
    def __init__(self, decoder: Callable[[str], JsonTypes], encoder: Callable[[JsonTypes], str]):
        super().__init__()
        self._decoder = decoder
        self._encoder = encoder
    
    def decode_entry(self,
                     subject: ObjectField,
                     object_id: int,
                     param1: str or None,  # old value
                     param2: str) -> ModificationEntry:  # new value
        return ModificationEntry(
            subject,
            object_id,
            None if param1 is None else self._decoder(param1),
            self._decoder(param2),
            param1 is None  # It's likely at creation if old value is None, but we can't be sure
        )
    
    def supports_encoding(self) -> bool:
        return True
    
    def encode_entry(self, entry: ModificationEntry) -> tuple[int, str or None, str]:
        return (entry.object_id,
                None if entry.is_creation else self._encoder(entry.value_old_json),
                self._encoder(entry.value_new_json))


class _DeletionEntryCodec(ChangelogEntryCodec[SimpleChangelogEntrySubject]):
    
    def decode_entry(self,
                     subject: SimpleChangelogEntrySubject,
                     object_id: int,
                     param1: str or None,
                     param2: str) -> DeletionEntry:
        return DeletionEntry(
            subject,
            object_id,
            _decode_legacy_boolean(param2)
        )
    
    def supports_encoding(self) -> bool:
        return True
    
    def encode_entry(self, entry: DeletionEntry) -> tuple[int, str or None, str]:
        return (entry.object_id,
                _encode_legacy_boolean(not entry.is_now_deleted),
                _encode_legacy_boolean(entry.is_now_deleted))


_CODEC_BOOLEAN = _LegacyModificationEntryCodec(_decode_legacy_boolean, _encode_legacy_boolean)
_CODEC_STRING_SHORT = _LegacyModificationEntryCodec(_decode_legacy_shortstring_to_string_short, _encode_identity)
_CODEC_STRING_TO_ID_STRING = _LegacyModificationEntryCodec(_decode_legacy_string_to_id_string, _encode_identity)
_CODEC_STRING_TO_SEMESTER_STRING = _LegacyModificationEntryCodec(_decode_legacy_string_to_semester_string, _encode_legacy_semester_string)
_CODEC_TEXT_TO_STRING_LONG = _LegacyModificationEntryCodec(_decode_legacy_text_to_string_long, _encode_identity)
_CODEC_DATETIME = _LegacyModificationEntryCodec(_decode_legacy_datetime, _encode_legacy_datetime)
_CODEC_INTEGER_NON_NEGATIVE = _LegacyModificationEntryCodec(_decode_legacy_integer_non_negative, _encode_legacy_integer)
_CODEC_LEGACY_ANNOUNCEMENT_LEVEL = _LegacyModificationEntryCodec(_decode_legacy_announcement_level, _encode_legacy_announcement_level)

_CODEC_DELETED = _DeletionEntryCodec()


def changelog_v1_build(builder: ChangelogVersionBuilder):
    builder.start_version(1)
    builder.set_object_class_id(CLASS_COURSE, "courses")
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("is_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("is_listed"), "listed", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("full_name"), "title", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("short_name"), "short", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("id_string"), "handle", _CODEC_STRING_TO_ID_STRING)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("organizer"), "organizer", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("topic"), "subject", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("semester"), "semester", _CODEC_STRING_TO_SEMESTER_STRING)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("allow_download"), "downloadable", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("internal_comment"), "internal", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_deletion_subject(), "deleted", _CODEC_DELETED)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("description"), "description", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("external"), "external", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("show_chapters_on_course"), "coursechapters", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("automatically_create_media_source_for_encoded_file"), "autopublish", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("make_new_videos_visible_by_default"), "autovisible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_COURSE.get_field("authentication_information"), "login_info", _CODEC_TEXT_TO_STRING_LONG)
    
    builder.set_object_class_id(CLASS_LECTURE, "lectures")
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("is_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("title"), "title", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("description"), "comment", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("internal_comment"), "internal", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("speaker"), "speaker", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("location"), "place", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("time"), "time", _CODEC_DATETIME)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("duration"), "duration", _CODEC_INTEGER_NON_NEGATIVE)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_deletion_subject(), "deleted", _CODEC_DELETED)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("livestream_planned"), "live", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_LECTURE.get_field("no_recording"), "norecording", _CODEC_BOOLEAN)
    # builder.add_mapping("lectures_data", "stream_settings", CLASS_LECTURE.get_field("authentication_information"), _convert_legacy_text_to_string_long)
    
    builder.set_object_class_id(CLASS_MEDIA_SOURCE, "videos")
    builder.set_subject_id_and_codec(CLASS_MEDIA_SOURCE.get_field("is_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_MEDIA_SOURCE.get_deletion_subject(), "deleted", _CODEC_DELETED)
    
    builder.set_object_class_id(CLASS_CHAPTER, "chapters")
    builder.set_subject_id_and_codec(CLASS_CHAPTER.get_field("start_time"), "time", _CODEC_INTEGER_NON_NEGATIVE)
    builder.set_subject_id_and_codec(CLASS_CHAPTER.get_field("name"), "text", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_CHAPTER.get_field("is_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_CHAPTER.get_deletion_subject(), "deleted", _CODEC_DELETED)
    
    builder.set_object_class_id(CLASS_ANNOUNCEMENT, "announcements")
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_field("text"), "text", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_field("type"), "level", _CODEC_LEGACY_ANNOUNCEMENT_LEVEL)
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_field("may_be_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_deletion_subject(), "deleted", _CODEC_DELETED)
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_field("publish_time"), "time_publish", _CODEC_DATETIME)
    builder.set_subject_id_and_codec(CLASS_ANNOUNCEMENT.get_field("expiration_time"), "time_expire", _CODEC_DATETIME)
    
    builder.set_object_class_id(CLASS_FEATURED, "featured")
    builder.set_subject_id_and_codec(CLASS_FEATURED.get_field("title"), "title", _CODEC_STRING_SHORT)
    builder.set_subject_id_and_codec(CLASS_FEATURED.get_field("text"), "text", _CODEC_TEXT_TO_STRING_LONG)
    builder.set_subject_id_and_codec(CLASS_FEATURED.get_field("is_visible"), "visible", _CODEC_BOOLEAN)
    builder.set_subject_id_and_codec(CLASS_FEATURED.get_deletion_subject(), "deleted", _CODEC_DELETED)
    
    builder.set_object_class_id(CLASS_PERMISSION, "perm")
    builder.set_subject_id_and_codec(CLASS_PERMISSION.get_deletion_subject(), "deleted", _CODEC_DELETED)
    
    builder.finish_version()
