from abc import ABC, ABCMeta, abstractmethod

from api.miscellaneous import *
from api.database import *
from api.objects.changelog import SimpleChangelogEntrySubject, ChangelogEntrySubject
from api.objects.changelog_entries import DeletionEntry, CreationEntry
from api.objects.field import ObjectField, ObjectColumnField, ObjectIndirectField
from api.objects.type import AbstractStringType


class ConfigurationExtra(ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self):
        super().__init__()
    
    @abstractmethod
    def queue_get_current_statements(self, transaction: ReadTransaction, object_id: int) -> tuple[FilledStatement, ...]:
        pass  # pragma: no cover
    
    @abstractmethod
    def get_current_from_results(self,
                                 transaction: ReadTransaction,
                                 statements: tuple[FilledStatement, ...]) -> tuple[str, JsonTypes]:
        pass  # pragma: no cover


class ObjectClass:
    
    def __init__(self,
                 api_id: str,
                 fields: list[ObjectField],
                 allow_get_config: bool,
                 allow_creation: bool,
                 allow_deletion: bool,
                 db_table: str,
                 variant_id_type: AbstractStringType or None = None,
                 variant_db_column: str or None = None,
                 variant_fields: dict[str, list[ObjectColumnField]] or None = None,
                 parent_classes: list[tuple["ObjectClass", str]] or None = None,
                 config_extras: list[ConfigurationExtra] or None = None):
        """
        :param fields: ID for API
        :param allow_get_config: If true, object configuration may be retrieved with the API
        :param allow_creation: If true, objects may be created with the API
        :param allow_deletion: If true, the object may be deleted with the API
        :param db_table: Database id of table
        :param variant_id_type: Type for object variant
        :param variant_db_column: Column where variant is stored
        :param variant_fields: Key is id of variant (according to variant_id_type)
        :param parent_classes: String is id of column where the parent id is stored
        :param config_extras: Extra entries to put in the configuration
        """
        
        self.api_id = api_id
        self._allow_get_config = allow_get_config
        self._allow_deletion = allow_deletion
        self._db_table = db_table
        self._variant_id_type = variant_id_type
        self._variant_db_column = variant_db_column
        self._variants_dict: dict[str, None] = None if variant_fields is None else \
            {key: None for key in variant_fields.keys()}
        self._parent_classes = None if parent_classes is None else {tup[0].api_id: tup for tup in parent_classes}
        self._config_extras = [] if config_extras is None else config_extras
        
        if parent_classes is not None and len(parent_classes) == 0:
            raise ValueError("Parent list may not be empty (Only None)")  # pragma: no cover
        if config_extras is not None and len(config_extras) == 0:
            raise ValueError("Config extras list may not be empty (Only None)")  # pragma: no cover
        
        if ((variant_id_type is None) != (variant_db_column is None)
                or (variant_id_type is None) != (variant_fields is None)):
            raise TypeError("variant_id_type, variant_db_column and variant_fields must all be set or all None")  # pragma: no cover

        db_ensure_valid_id(db_table)
        used_column_ids: dict[str, None] = {}
        
        if variant_db_column is not None:
            db_ensure_valid_id(variant_db_column)
            if variant_db_column in used_column_ids:
                raise ValueError(f"Duplicated column id {variant_db_column}")  # pragma: no cover
            used_column_ids[variant_db_column] = None
        
        if parent_classes is not None:
            for parent_class, column in parent_classes:
                if column in used_column_ids:
                    raise ValueError(f"Duplicated column id {column}")  # pragma: no cover
                used_column_ids[column] = None
        
        self._object_query_request = PreparedStatement(f"""
            SELECT *
            FROM "{db_table}"
            WHERE (NOT "deleted")
                AND "id" = ?
            """)
        
        self._object_query_request_no_deletion_check = PreparedStatement(f"""
            SELECT *
            FROM "{db_table}"
            WHERE "id" = ?
            """)
        
        # Key is field id. First tuple value is variant id
        self._all_fields: dict[str, tuple[str or None, ObjectField]] = {}
        self._indirect_fields: list[ObjectIndirectField] = []
        
        for field in fields:
            if field.api_id in self._all_fields:
                raise ValueError(f"Duplicated field id {field.api_id}")  # pragma: no cover
            field.on_added_to_class(self)
            assert isinstance(field, ObjectColumnField) or isinstance(field, ObjectIndirectField)
            self._all_fields[field.api_id] = (None, field)
            if isinstance(field, ObjectColumnField):
                if field.db_column in used_column_ids:
                    raise ValueError(f"Duplicated column id {field.db_column}")  # pragma: no cover
                used_column_ids[field.db_column] = None
            else:
                assert isinstance(field, ObjectIndirectField)
                self._indirect_fields.append(field)
        
        if variant_fields is not None:
            for variant_id in variant_fields.keys():
                variant_used_column_ids = used_column_ids.copy()
                for field in variant_fields[variant_id]:
                    if field.api_id in self._all_fields:
                        raise ValueError(f"Duplicated field id {field.api_id}")  # pragma: no cover
                    if field.db_column in variant_used_column_ids:  # pragma: no cover
                        raise ValueError(f"Duplicated column id {field.db_column} (Duplicates for different variants "
                                         f"are allowed)")
                    field.on_added_to_class(self)
                    variant_used_column_ids[field.db_column] = None
                    self._all_fields[field.api_id] = (variant_id, field)

        self._deletion_subject = SimpleChangelogEntrySubject(self)
        self._creation_subject = SimpleChangelogEntrySubject(self)
        self._all_log_subjects = ([self._deletion_subject, self._creation_subject] +
                                  list(map(lambda t: t[1], self._all_fields.values())))
        
        self._creation_config_json = None
        if allow_creation:
            self._creation_config_json = {
                "fields": [field.get_description() for field in fields if field.include_in_config]
            }
            if variant_fields is not None:
                assert self._variant_id_type is not None
                variant_config_list = {}
                self._creation_config_json["variant_fields"] = variant_config_list
                for variant_id in variant_fields.keys():
                    variant_config_list[variant_id] = [field.get_description()
                                                       for field in variant_fields[variant_id]
                                                       if field.include_in_config]
    
    def legacy_get_db_table(self) -> str:
        return self._db_table
    
    def get_deletion_subject(self) -> SimpleChangelogEntrySubject:
        return self._deletion_subject
    
    def get_creation_subject(self) -> SimpleChangelogEntrySubject:
        return self._creation_subject
    
    def get_all_log_subjects(self) -> list[ChangelogEntrySubject]:
        return self._all_log_subjects
    
    def get_all_field_ids(self):
        return self._all_fields.keys()
    
    def get_field(self, field_id: str) -> ObjectField:
        return self._all_fields[field_id][1]
    
    def is_get_config_allowed(self) -> bool:
        return self._allow_get_config
    
    def _db_execute_get_config(self, transaction: ReadTransaction, object_id: int) -> dict:
        indirect_statements: dict[ObjectIndirectField, tuple[FilledStatement, ...]] = {}
        config_extra_statements: list[tuple[ConfigurationExtra, tuple[FilledStatement, ...]]] = []
        for field in self._indirect_fields:
            if field.include_in_config:
                indirect_statements[field] = field.queue_get_current_statements(transaction, object_id)
        for extra in self._config_extras:
            config_extra_statements.append((extra, extra.queue_get_current_statements(transaction, object_id)))
        object_db_list = transaction.execute_statement_and_close(
            self._object_query_request, object_id
        )
        
        if len(object_db_list) < 1:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        if len(object_db_list) > 1:
            raise ValueError("Got more than one object with id " + str(object_id))  # pragma: no cover
        object_db = object_db_list[0]
        
        configurations = []
        config = {
            "fields": configurations
        }
        
        object_variant_json = None
        if self._variant_id_type is not None:
            object_variant_json = self._variant_id_type.db_value_to_json(object_db[self._variant_db_column])
            assert isinstance(object_variant_json, str)
        
        for variant_id, field in self._all_fields.values():
            if variant_id is not None and variant_id != object_variant_json:
                continue
            if not field.include_in_config:
                continue
            if isinstance(field, ObjectColumnField):
                current_value = field.get_current_value_from_db(object_db)
            else:
                assert isinstance(field, ObjectIndirectField)
                current_value = field.get_json_value_from_results(
                    transaction,
                    object_db,
                    indirect_statements[field]
                )
            configurations.append(field.get_value_json(current_value))
        
        for extra, stats in config_extra_statements:
            key, value = extra.get_current_from_results(transaction, stats)
            config[key] = value
        
        return config
    
    def get_current_config(self, object_id: int) -> JsonTypes or None:
        if not self._allow_get_config:
            return None
        
        return db_pool.execute_read_transaction(self._db_execute_get_config, object_id)
    
    def modify_current_config(self,
                              transaction: WriteTransaction,
                              modifying_user_id: int,
                              object_id: int,
                              expected_old_values: CJsonObject,
                              new_values: CJsonObject):
        
        column_updates: list[tuple[str, DbValueType]] = []
        indirect_updates: list[tuple[ObjectIndirectField, CJsonValue, object]] = []
        from api.objects.current_changelog import UpdateChangelogUpdate
        changelog_update: UpdateChangelogUpdate = UpdateChangelogUpdate()
        
        object_list_db = transaction.execute_statement(self._object_query_request, object_id)
        if len(object_list_db) == 0:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        object_db = object_list_db[0]
        
        object_variant_json = None
        if self._variant_id_type is not None:
            object_variant_json = self._variant_id_type.db_value_to_json(object_db[self._variant_db_column])
            assert isinstance(object_variant_json, str)
        
        for field_id in new_values.keys():
            if field_id not in self._all_fields:
                raise ApiClientException(ERROR_OBJECT_ERROR(f"Unknown field '{truncate_string(field_id)}'"))
            variant_id, field = self._all_fields[field_id]
            if variant_id is not None:
                if object_variant_json != variant_id:
                    raise ApiClientException(ERROR_OBJECT_ERROR(
                        f"Field '{truncate_string(field_id)}' is for other variant"))
            
            new_value = new_values.get(field.api_id)
            expected_old_value = None
            if expected_old_values.has(field_id):
                expected_old_value = expected_old_values.get(field_id)
            
            if isinstance(field, ObjectColumnField):
                update = field.do_value_update(
                    transaction, changelog_update, object_id, object_db, expected_old_value, new_value)
                if update is not None:
                    column_updates.append(update)
            else:
                assert isinstance(field, ObjectIndirectField)
                indirect_updates.append((
                    field,
                    expected_old_value,
                    field.validate_and_convert_client_value(transaction, object_id, new_value)
                ))
        
        if len(column_updates) > 0:
            transaction.queue_statement(f"""
                UPDATE "{self._db_table}"
                SET {",".join(map(lambda t: f'"{t[0]}"=?', column_updates))}
                WHERE "id" = ?
                """, *[new_value_db for _, new_value_db in column_updates], object_id)
        
        for field, expected_old_value, value in indirect_updates:
            field.queue_update(transaction, changelog_update, object_id, expected_old_value, value)
        
        changelog_update.queue_log_statements(transaction, modifying_user_id)
    
    def is_creation_allowed(self) -> bool:
        return self._creation_config_json is not None
    
    def is_deletion_allowed(self) -> bool:
        return self._allow_deletion
    
    def get_creation_config(self) -> JsonTypes or None:
        return self._creation_config_json
    
    def create(self,
               modifying_user_id: int,
               parent_type: str or None,
               parent_id: int or None,
               values: CJsonObject,
               object_variant_client: CJsonValue or None) -> int:
        if self._creation_config_json is None:
            raise RuntimeError("Cannot create objects of this type")  # pragma: no cover
        
        parent_class: ObjectClass or None = None
        parent_id_column: int or None = None
        if self._parent_classes is not None:
            if parent_type is None:
                raise ApiClientException(ERROR_OBJECT_ERROR("Missing parent type"))
            if parent_id is None:
                raise ApiClientException(ERROR_OBJECT_ERROR("Missing parent id"))
            if parent_type not in self._parent_classes:
                raise ApiClientException(ERROR_OBJECT_ERROR("Unknown parent type"))
            parent_class, parent_id_column = self._parent_classes[parent_type]
        else:
            if parent_type is not None or parent_id is not None:
                raise ApiClientException(ERROR_OBJECT_ERROR("This object may have no parent"))
        
        if self._variant_id_type is None and object_variant_client is not None:
            raise ApiClientException(ERROR_OBJECT_ERROR("This object has no variant"))
        if self._variant_id_type is not None and object_variant_client is None:
            raise ApiClientException(ERROR_OBJECT_ERROR("Missing object variant"))
        
        return db_pool.execute_write_transaction(
            self._db_execute_create,
            modifying_user_id,
            parent_id,
            values,
            object_variant_client,
            parent_class,
            parent_id_column
        )
    
    def _db_execute_create(self,
                           transaction: WriteTransaction,
                           modifying_user_id: int,
                           parent_id: int or None,
                           values: CJsonObject,
                           object_variant_client: CJsonValue or None,
                           parent_class: "ObjectClass" or None,
                           parent_id_column: int or None) -> int:
        from api.objects.current_changelog import CreationChangelogUpdate
        changelog_update: CreationChangelogUpdate = CreationChangelogUpdate()
        
        object_variant_db = None
        object_variant_json: str or None = None
        if self._variant_id_type is not None:
            object_variant_json = object_variant_client.as_string(self._variant_id_type.get_max_string_length())
            if object_variant_json not in self._variants_dict:
                raise ApiClientException(ERROR_OBJECT_ERROR(
                    f"Unknown variant {truncate_string(object_variant_json)}"))
            try:
                object_variant_db = self._variant_id_type.validate_and_convert_client_value(
                    transaction, None, object_variant_client)
            except Exception as e:  # pragma: no cover
                raise RuntimeError(f"Variant {truncate_string(object_variant_json)} is declared as possible "
                                   f"variant but variant type does not accept it") from e
        
        column_updates: list[tuple[str, int]] = []
        indirect_updates: list[tuple[ObjectIndirectField, object]] = []
        for variant_id, field in self._all_fields.values():
            if not values.has(field.api_id):
                if not field.include_in_config:
                    continue
                if variant_id is not None and variant_id != object_variant_json:
                    continue
                if field.default_value_json is None:
                    raise ApiClientException(ERROR_OBJECT_ERROR(f"Missing field {truncate_string(field.api_id)}"))
                field_value = CJsonValue(field.default_value_json if field.default_value_json != (None,) else None)
            else:
                if not field.include_in_config:
                    raise ApiClientException(ERROR_OBJECT_ERROR(
                        f"Field {truncate_string(field.api_id)} may not be included in creation"))
                if variant_id is not None and variant_id != object_variant_json:
                    raise ApiClientException(ERROR_OBJECT_ERROR(f"Field {truncate_string(field.api_id)} is for other "
                                                                f"variant"))
                field_value = values.get(field.api_id)
            
            if isinstance(field, ObjectColumnField):
                update = field.do_value_creation(transaction, changelog_update, field_value)
                if update is not None:
                    column_updates.append(update)
            else:
                assert isinstance(field, ObjectIndirectField)
                indirect_updates.append((
                    field,
                    field.validate_and_convert_client_value(transaction, None, field_value)
                ))
        
        if parent_class is not None:
            assert isinstance(parent_class, ObjectClass)
            parent_list_db = transaction.execute_statement(parent_class._object_query_request, parent_id)
            if len(parent_list_db) == 0:
                raise ApiClientException(ERROR_OBJECT_ERROR("Parent does not exist"))
            column_updates.append((parent_id_column, parent_id))
        
        if object_variant_db is not None:
            column_updates.append((self._variant_db_column, object_variant_db))
        
        object_id = transaction.execute_statement(f"""
            INSERT INTO "{self._db_table}" ({",".join(map(lambda t: f'"{t[0]}"', column_updates))})
            VALUES ({"" if len(column_updates) == 0 else ("?," * (len(column_updates) - 1) + "?")})
            RETURNING "id"
            """, *[new_value_db for _, new_value_db in column_updates])[0]["id"]
        changelog_update.add_entry(CreationEntry(
            self._creation_subject,
            object_id,
            object_variant_json,
            parent_class,
            parent_id
        ))
        for field, value in indirect_updates:
            field.queue_creation(
                transaction,
                changelog_update,
                object_id,
                value
            )
        changelog_update.queue_log_statements(
            transaction, modifying_user_id, object_id)
        transaction.commit()
        return object_id
    
    def delete(self, modifying_user_id: int, object_id: int):
        db_pool.execute_write_transaction(self._db_execute_delete, modifying_user_id, object_id)
    
    def _db_execute_delete(self, transaction: WriteTransaction, modifying_user_id: int, object_id: int):
        if len(transaction.execute_statement(self._object_query_request, object_id)) == 0:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        from api.objects.current_changelog import UpdateChangelogUpdate
        changelog = UpdateChangelogUpdate()
        
        changelog.add_entry(DeletionEntry(
            self._deletion_subject,
            object_id,
            True
        ))
        transaction.queue_statement(f"""
            UPDATE "{self._db_table}"
            SET "deleted" = true
            WHERE "id" = ?
            """, object_id)
        changelog.queue_log_statements(transaction, modifying_user_id)
        transaction.commit()
    
    def undo_deletion(self, modifying_user_id: int, object_id: int):
        db_pool.execute_write_transaction(self._db_execute_undo_deletion, modifying_user_id, object_id)
    
    def _db_execute_undo_deletion(self, transaction: WriteTransaction, modifying_user_id: int, object_id: int):
        object_list_db = transaction.execute_statement(self._object_query_request_no_deletion_check, object_id)
        if len(object_list_db) == 0:
            raise ApiClientException(ERROR_UNKNOWN_OBJECT)
        if not bool(object_list_db[0]["deleted"]):
            raise ApiClientException(ERROR_OBJECT_ERROR("Object not deleted"))
        
        from api.objects.current_changelog import UpdateChangelogUpdate
        changelog = UpdateChangelogUpdate()
        
        changelog.add_entry(DeletionEntry(
            self._deletion_subject,
            object_id,
            False
        ))
        transaction.queue_statement(f"""
            UPDATE "{self._db_table}"
            SET "deleted" = false
            WHERE "id" = ?
            """, object_id)
        changelog.queue_log_statements(transaction, modifying_user_id)
        transaction.commit()
